from __future__ import absolute_import

from omegaml.store import qops
from omegaml.store.query import Filter
from omegaml.util import PickableCollection, ensure_base_collection


class FilteredCollection:
    """
    A permanently filtered collection

    Supports all methods as a Collection does, however any filter or query
    argument is permanently set at instantiation

        fcoll = FilteredCollection(collection, query={ expression })

    Any subsequent operation will automatically apply the query expression.

    Note that v.v. a Collection and all methods that accept a filter as their first
    argument have a changed signature - the filter argument is optional
    with all FilteredCollection methods, as the filter is set at instantiation.

        Example:

            # in pymongo

            filter = {expression}
            coll.find_one_and_replace(filter, replace)

            # FilteredCollection

            coll = FilteredCollection(query={expression})
            coll.find_one_and_replace(replace, filter=None)

    This is so that calls to a FilteredCollection feel more natural, as opposed
    to specifying an empty filter argument on every call. Still, an additional
    filter can be specified on every method that accepts the filter= optional
    argument:

            # temporarily add another filter

            coll.find_one_and_replace(replace, filter={expression})

    Here expression will only apply to this particular method call. The
    global filter set by query= is unchanged.

    If no expression is given, the empty expression {} is assumed. To change
    the expression for the set fcoll.query = { expression }
    """

    def __init__(self, collection, query=None, projection=None, **kwargs):
        if isinstance(collection, FilteredCollection):
            # avoid cascading of FilteredCollections
            query = query or collection._fixed_query
            projection = projection or collection.projection
            collection = ensure_base_collection(collection)
        else:
            query = query or {}
        self._fixed_query = query
        self.projection = projection
        self.collection = PickableCollection(collection)

    @property
    def _Collection__database(self):
        return self.collection.database

    @property
    def name(self):
        return self.collection.name

    @property
    def database(self):
        return self.collection.database

    @property
    def query(self):
        return Filter(self.collection, **self._fixed_query).query

    def aggregate(self, pipeline, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        pipeline.insert(0, qops.MATCH(query))
        kwargs.update(allowDiskUse=True)
        return self.collection.aggregate(pipeline, **kwargs)

    def find(self, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.find(filter=query, **kwargs)

    def find_one(self, filter=None, *args, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.find_one(query, *args, **kwargs)

    def find_one_and_delete(self, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.find_one_and_delete(query,
                                                   **kwargs)

    def find_one_and_replace(self, replacement, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.find_one_and_replace(query,
                                                    replacement,
                                                    **kwargs)

    def find_one_and_update(self, update, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.find_one_and_update(query,
                                                   update,
                                                   **kwargs)

    def count(self, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.count(filter=query, **kwargs)

    def distinct(self, key, filter=None, **kwargs):
        query = dict(self.query)
        query.update(filter or {})
        return self.collection.distinct(key, filter=query, **kwargs)

    def create_index(self, keys, **kwargs):
        return self.collection.create_index(keys, **kwargs)

    def list_indexes(self, **kwargs):
        return self.list_indexes(**kwargs)

    def group(self, key, initial, reduce, condition=None, **kwargs):
        condition = dict(self.query)
        condition.update(condition or {})
        return self.collection.group(key, condition, initial,
                                     reduce,
                                     **kwargs)

    def map_reduce(self, m, r, out, full_response=False, query=None, **kwargs):
        _query = dict(self.query)
        _query.update(query or {})
        return self.collection.map_reduce(m, r, out,
                                          full_response=False,
                                          query=_query, **kwargs)

    def inline_map_reduce(self, m, r, full_response=False,
                          query=None, **kwargs):
        _query = dict(self.query)
        _query.update(query or {})
        return self.collection.inline_map_reduce(m, r,
                                                 full_response=False, query=_query, **kwargs)

    def insert(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")

    def update(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")

    def remove(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")

    def find_and_modify(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")

    def ensure_index(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")

    def save(self, *args, **kwargs):
        raise NotImplementedError(
            "deprecated in Collection and not implemented in FilteredCollection")



from unittest import TestCase

from shutil import rmtree

from omegaml import Omega
from omegaml.backends.package import PythonPipSourcedPackageData
from omegaml.util import settings, delete_database, mkdirs


class PythonRemotePackageDataTests(TestCase):
    _multiprocess_can_split_ = False

    def setUp(self):
        self.defaults = settings()
        OMEGA_STORE_BACKENDS = self.defaults.OMEGA_STORE_BACKENDS
        self.backend = backend = 'omegaml.backends.package.PythonPipSourcedPackageData'
        if PythonPipSourcedPackageData.KIND not in OMEGA_STORE_BACKENDS:
            OMEGA_STORE_BACKENDS[PythonPipSourcedPackageData.KIND] = backend
        self.om = Omega()
        delete_database()
        self.pkgsdir = self.om.scripts.get_backend_bykind(PythonPipSourcedPackageData.KIND).packages_path
        mkdirs(self.pkgsdir)

    def tearDown(self):
        rmtree(self.pkgsdir, ignore_errors=True)

    def test_put_pypi(self):
        om = self.om
        pkgsrc = 'pypi://six'
        om.scripts.put(pkgsrc, 'six')
        metas = om.scripts.list(raw=True)
        self.assertEqual(metas[0].kind, 'pipsrc.package')

    def test_get_pypi(self):
        om = self.om
        pkgsrc = 'pypi://six'
        om.scripts.put(pkgsrc, 'six')
        mod1 = om.scripts.get('six')
        self.assertIsNotNone(mod1)
        self.assertIn('six', mod1.__file__)

    def test_get_pypi_versioned(self):
        om = self.om
        pkgsrc = 'pypi://six==1.15.0'
        om.scripts.put(pkgsrc, 'six')
        mod1 = om.scripts.get('six')
        self.assertIsNotNone(mod1)
        self.assertIn('six', mod1.__file__)

    def test_get_git(self):
        om = self.om
        pkgsrc = 'git+https://github.com/benjaminp/six.git'
        om.scripts.put(pkgsrc, 'six')
        mod1 = om.scripts.get('six')
        self.assertIsNotNone(mod1)
        self.assertIn('six', mod1.__file__)

    def test_get_git_versioned(self):
        om = self.om
        pkgsrc = 'git+https://github.com/benjaminp/six.git@1.15.0'
        om.scripts.put(pkgsrc, 'six')
        mod1 = om.scripts.get('six')
        self.assertIsNotNone(mod1)
        self.assertIn('six', mod1.__file__)



from unittest import TestCase

import numpy as np
from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.mixins.store.modelversion import ModelVersionMixin
from omegaml.tests.util import OmegaTestMixin


class ModelVersionMixinTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()

    def test_version_on_put(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        clf = LinearRegression()
        meta = store.put(clf, 'regmodel')
        self.assertIn('versions', meta.attributes)
        models = store.list(include_temp=True)
        latest = meta.attributes['versions']['tags']['latest']
        store_key = store._model_version_store_key('regmodel', latest)
        self.assertIn(store_key, models)

    def test_get_version_by_index(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        clf = LinearRegression()
        clf.version_ = 1
        meta = store.put(clf, 'regmodel')
        clf.version_ = 2
        meta = store.put(clf, 'regmodel')
        clf_ = store.get('regmodel', version=-1)
        self.assertEqual(clf_.version_, 2)
        clf_ = store.get('regmodel', version=-2)
        self.assertEqual(clf_.version_, 1)
        clf_ = store.get('regmodel')
        self.assertEqual(clf_.version_, 2)

    def test_get_version_by_tag(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        clf = LinearRegression()
        clf.version_ = 1
        meta = store.put(clf, 'regmodel', tag='version1')
        clf.version_ = 2
        meta = store.put(clf, 'regmodel', tag='version2')
        clf_ = store.get('regmodel', tag='version2')
        self.assertEqual(clf_.version_, 2)
        clf_ = store.get('regmodel', tag='version1')
        self.assertEqual(clf_.version_, 1)
        clf_ = store.get('regmodel')
        self.assertEqual(clf_.version_, 2)

    def test_get_version_by_attag(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        clf = LinearRegression()
        clf.version_ = 1
        meta = store.put(clf, 'regmodel', tag='version1')
        clf.version_ = 2
        meta = store.put(clf, 'regmodel', tag='version2')
        clf_ = store.get('regmodel@version2')
        self.assertEqual(clf_.version_, 2)
        clf_ = store.get('regmodel@version1')
        self.assertEqual(clf_.version_, 1)
        clf_ = store.get('regmodel')
        self.assertEqual(clf_.version_, 2)

    def test_get_version_by_commit(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        clf = LinearRegression()
        clf.version_ = 1
        meta = store.put(clf, 'regmodel')
        clf.version_ = 2
        meta = store.put(clf, 'regmodel')
        meta = store.metadata('regmodel')
        commit1 = meta.attributes['versions']['commits'][-2]['ref']
        commit2 = meta.attributes['versions']['commits'][-1]['ref']
        clf_ = store.get('regmodel', commit=commit2)
        self.assertEqual(clf_.version_, 2)
        clf_ = store.get('regmodel', commit=commit1)
        self.assertEqual(clf_.version_, 1)
        clf_ = store.get('regmodel')
        self.assertEqual(clf_.version_, 2)

    def test_get_metadata_by_version(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        clf = LinearRegression()
        clf.version_ = 1
        meta1 = store.put(clf, 'regmodel', tag='commit1')
        clf.version_ = 2
        meta2 = store.put(clf, 'regmodel', tag='commit2')
        self.assertEqual(meta1.id, meta2.id)
        meta_commit1_byname = store.metadata(meta2.attributes['versions']['commits'][-2]['name'])
        meta_commit2_byname = store.metadata(meta2.attributes['versions']['commits'][-1]['name'])
        meta_commit1_bymeta = store.metadata('regmodel@commit1', raw=True)
        meta_commit2_bymeta = store.metadata('regmodel@commit2', raw=True)
        self.assertEqual(meta_commit1_bymeta.id, meta_commit1_byname.id)
        self.assertEqual(meta_commit2_bymeta.id, meta_commit2_byname.id)

    def test_via_runtime(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        reg = LinearRegression()
        reg.coef_ = np.array([2])
        reg.intercept_ = 10
        store.put(reg, 'regmodel', tag='commit1')
        reg.coef_ = np.array([5])
        reg.intercept_ = 0
        store.put(reg, 'regmodel', tag='commit2')
        # via past version pointer
        r1 = self.om.runtime.model('regmodel^').predict([10]).get()
        r2 = self.om.runtime.model('regmodel').predict([10]).get()
        self.assertEqual(r1[0], 10 * 2 + 10)
        self.assertEqual(r2[0], 10 * 5 + 0)
        # via version tag
        r1 = self.om.runtime.model('regmodel@commit1').predict([10]).get()
        r2 = self.om.runtime.model('regmodel@commit2').predict([10]).get()
        self.assertEqual(r1[0], 10 * 2 + 10)
        self.assertEqual(r2[0], 10 * 5 + 0)

    def test_nonexistent(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        store.metadata('nonexistent')
        store.get('nonexistent')

    def test_dropversion(self):
        store = self.om.models
        store.register_mixin(ModelVersionMixin)
        reg = LinearRegression()
        reg.coef_ = np.array([2])
        reg.intercept_ = 10
        store.put(reg, 'regmodel', tag='commit1')
        reg.coef_ = np.array([5])
        reg.intercept_ = 0
        store.put(reg, 'regmodel', tag='commit2')







import base64
from unittest import TestCase

import json
from nbformat import NotebookNode
from tornado.web import HTTPError

from omegaml import Omega
from omegaml.notebook.omegacontentsmgr import OmegaStoreContentsManager


class OmegaContentsManagerTests(TestCase):
    def setUp(self):
        self.om = Omega()
        self.mgr = OmegaStoreContentsManager(omega=self.om)
        [self.om.jobs.drop(fn) for fn in self.om.jobs.list()]
        [self.om.datasets.drop(fn) for fn in self.om.datasets.list()]

    def _create_notebook(self, name):
        code = """
        print('hello world')
        """.strip()
        self.om.jobs.create(code, name)

    def test_get_top_level(self):
        om = self.om
        # check director is empty to start with
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 0)
        # add a new model, check it is represented
        self._create_notebook('foo')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        self.assertEqual(model['content'][0]['type'], 'notebook')

    def test_subdirectory(self):
        om = self.om
        # create a notebook in a sub directory
        self._create_notebook('sub/foo')
        model = self.mgr.get('/')
        # check we can see the directory
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['type'], 'directory')
        self.assertEqual(model['content'][0]['name'], 'sub')
        # check we can see the notebook
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        # create a notebook in a sub sub directory
        self._create_notebook('sub1/sub2/foo')
        model = self.mgr.get('/sub1', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['type'], 'directory')
        self.assertEqual(model['content'][0]['name'], 'sub2')
        model = self.mgr.get('/sub1/sub2', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')

    def test_save_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        # copy foo to bar
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr._notebook_model('foo.ipynb')
        self.mgr.save(model, 'sub/bar.ipynb')
        # check it got created within a directory
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'bar.ipynb')
        # check it can be read directly
        model = self.mgr.get('/sub/bar.ipynb', type='notebook')
        self.assertEqual(model['name'], 'bar.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world')")

    def test_blankfile_exists_always_false(self):
        # this protects against cannot use POST, use PUT error
        # see https://stackoverflow.com/a/59775912/890242
        # create a first notebook
        result = self.mgr.file_exists('')
        self.assertEqual(result, False)
        self._create_notebook('foo')
        result = self.mgr.file_exists('')
        self.assertEqual(result, False)
        result = self.mgr.file_exists('foo2')
        self.assertEqual(result, False)
        self._create_notebook('sub/foo')
        result = self.mgr.file_exists('sub')
        self.assertEqual(result, False)

    def test_get_notebook_as_file(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        # check it can be read directly as a file
        model = self.mgr.get('foo.ipynb', type='file')
        nb_content = json.loads(model['content'])
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(nb_content['cells'][0]['source'], ["print('hello world')"])

    def test_save_notebook_as_file(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        # check it can be read directly as a file
        model = self.mgr.get('foo.ipynb', type='file')
        model['name'] = 'bar.ipynb'
        self.mgr.save(model, 'bar.ipynb')
        self.assertIn('bar.ipynb', self.om.jobs.list())
        nb = self.om.jobs.get('bar.ipynb')
        self.assertIsInstance(nb, NotebookNode)

    def test_save_notebook_as_file_base64(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        # check it can be read directly as a file
        model = self.mgr.get('foo.ipynb', type='file', format='base64')
        model['name'] = 'bar.ipynb'
        model['format'] = 'base64'
        model['content'] = base64.encodebytes(json.dumps(model['content']).encode('utf8'))
        self.mgr.save(model, 'bar.ipynb')
        self.assertIn('bar.ipynb', self.om.jobs.list())
        nb = self.om.jobs.get('bar.ipynb')
        self.assertIsInstance(nb, NotebookNode)

    def test_overwrite_existing_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        # copy foo to bar
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr._notebook_model('foo.ipynb')
        self.mgr.save(model, 'sub/foo.ipynb')
        # save again
        model['content']['cells'][0]['source'] = "print('hello world at large!')"
        self.mgr.save(model, 'sub/foo.ipynb')
        # check it got created within a directory
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'foo.ipynb')
        # check it can be read directly
        model = self.mgr.get('/sub/foo.ipynb', type='notebook')
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world at large!')")

    def test_create_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test creating directory on sub level
        model = self.mgr._base_model('sub/sub1', kind='directory')
        self.mgr.save(model, 'sub/sub1')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        model = self.mgr.get('/sub', type='directory')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub1')
        self.assertEqual(model['content'][0]['type'], 'directory')

    def test_get_directory_contents(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test getting empty directory
        model = self.mgr.get('/sub')
        self.assertEqual('sub', model['name'])
        self.assertEqual('directory', model['type'])
        self.assertEqual([], model['content'])

    def test_create_directory_escaped(self):
        # test creating directory on top level
        model = self.mgr._base_model('Untitled Folder', kind='directory')
        self.mgr.save(model, 'Untitled Folder')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'Untitled Folder')
        self.assertEqual(model['content'][0]['type'], 'directory')
        # test getting it with escaped string
        self.assertTrue(self.mgr.exists('Untitled Folder'))
        self.assertTrue(self.mgr.exists('Untitled%20Folder'))

    def test_create_notebook_in_empty_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        result = self.mgr.exists('sub')
        self.assertTrue(result)
        # test creating a notebook in this new emtpty directory
        self._create_notebook('sub/foo.ipynb')
        model = self.mgr.get('sub/foo.ipynb', type='notebook')
        self.assertEqual(model['name'], 'foo.ipynb')
        self.assertEqual(model['content']['cells'][0]['source'], "print('hello world')")

    def test_create_multiple_notebook_in_empty_directory(self):
        # test creating directory on top level
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')
        result = self.mgr.exists('sub')
        self.assertTrue(result)
        # test creating a notebook in this new emtpty directory
        self._create_notebook('sub/foo.ipynb')
        self._create_notebook('sub/foo2.ipynb')
        self._create_notebook('sub/foo3.ipynb')
        model = self.mgr.get('/')
        self.assertEqual(len(model['content']), 1)
        self.assertEqual(model['content'][0]['name'], 'sub')
        self.assertEqual(model['content'][0]['type'], 'directory')

    def test_exists(self):
        om = self.om
        # check directory is empty to start with
        result = self.mgr.exists('')
        self.assertTrue(result)
        result = self.mgr.dir_exists('')
        self.assertTrue(result)
        result = self.mgr.file_exists('')
        self.assertFalse(result)
        result = self.mgr.file_exists('test.ipynb')
        self.assertFalse(result)
        # check directory is empty to start with
        result = self.mgr.exists('/Untitled1')
        self.assertFalse(result)
        result = self.mgr.exists('/sub/sub1')
        self.assertFalse(result)
        # test creating directory on sub level
        model = self.mgr._base_model('sub/sub1', kind='directory')
        self.mgr.save(model, 'sub/sub1')
        result = self.mgr.exists('sub/sub1')
        self.assertTrue(result)
        # test dropping the directory actually works
        self.mgr.delete('sub/sub1')
        result = self.mgr.exists('sub/sub1')
        self.assertFalse(result)
        # test creating a file works
        model = self.mgr._base_model('Notebook.ipynb', kind='notebook')
        result = self.mgr.exists('Notebook.ipynb')

    def test_rename_notebook(self):
        # create a dummy notebook just so we have a valid model
        self._create_notebook('foo')
        nbmodel = self.mgr.get('foo.ipynb', type='notebook')
        self.mgr.rename('foo.ipynb', 'bar.ipynb')

    def test_rename_directory(self):
        model = self.mgr._base_model('sub', kind='directory')
        self.mgr.save(model, 'sub')
        self.mgr.rename_file('sub', 'new_sub')
        with self.assertRaises(HTTPError):
            model = self.mgr.get('sub', type='directory')
        model = self.mgr.get('new_sub', type='directory')
        self.assertEqual(len(model['content']), 0)

    def test_save_file_base64(self):
        model = self.mgr._base_model('textfile.txt', kind='file')
        model['content'] = base64.encodebytes('hello world'.encode('utf8')).decode('ascii')
        model['format'] = 'base64'
        self.mgr.save(model, 'textfile.txt')
        model = self.mgr.get('textfile.txt', type='file')
        self.assertEqual(model['content'], 'hello world')

    def test_save_file_text(self):
        model = self.mgr._base_model('textfile.txt', kind='file')
        model['content'] = 'hello world'
        model['format'] = 'text'
        self.mgr.save(model, 'textfile.txt')
        model = self.mgr.get('textfile.txt', type='file')
        self.assertEqual(model['content'], 'hello world')

    def test_weird_filenames(self):
        for fn in ('Untitled (1).ipynb', 'Some+strange+name.+++'):
            self._create_notebook(fn)
            expected_fn = fn + ('.ipynb' if not fn.endswith('.ipynb') else '')
            self.assertIn(expected_fn, self.om.jobs.list())
            model = self.mgr.get('/')
            contents = [e['name'] for e in model['content']]
            self.assertIn(expected_fn, contents)



from pymongo import MongoClient as RealMongoClient


def MongoClient(*args, **kwargs):
    """
    Shim function adding SSL kwargs on MongoClient instead of changing
    each and every location
    """
    from omegaml import settings
    defaults = settings()
    kwargs.update(defaults.OMEGA_MONGO_SSL_KWARGS)
    return RealMongoClient(*args, **sanitize_mongo_kwargs(kwargs))


def sanitize_mongo_kwargs(kwargs):
    # keep kwargs sane
    # -- if we receive "use_ssl" = False from cloud config but have
    #    a local CA defined, remove it. Otherwise mongo raises ConfigurationError
    kwargs = dict(kwargs)
    if 'ssl_ca_certs' in kwargs and not kwargs.get('ssl'):
        del kwargs['ssl_ca_certs']
    return kwargs



class PythonPackageMixin(object):
    """
    Install and load scripts
    """
    def install(self, specs=None):
        """
        install and load packages

        Args:
            specs (str, list): optional, package name or list of names,
                defaults to self.list()
        """
        specs = specs or self.list()
        if isinstance(specs, str):
            specs = specs.split(' ')
        for pkgname in specs:
            self.get(pkgname)



import os
from unittest import TestCase

import sys

from omegaml import Omega
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.restapi.app import app
from omegaml.restapi.tests.util import RequestsLikeTestClient
from omegaml.tests.util import OmegaTestMixin


class ScriptResourceTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.client = RequestsLikeTestClient(app)
        self.om = Omega()
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()

    def tearDown(self):
        pass

    def url(self, pk=None, action=None, query=None):
        url = '/api/v1/script/'
        if pk is not None:
            url += '{pk}/'.format(**locals())
        if action is not None:
            url += '{action}'.format(**locals())
        if query is not None:
            url += '?{query}'.format(**locals())
        return url

    def test_script_run(self):
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        # put script
        meta = om.scripts.put(pkg, 'helloworld')
        # run the script on the cluster
        resp = self.client.post(self.url('helloworld', action='run', query='text=foo'))
        self.assertHttpOK(resp)
        data = self.deserialize(resp)
        self.assertIn('runtimes', data)
        self.assertIn('result', data)
        expected = list(['hello from helloworld', {'text': 'foo', 'pure_python': False}])
        self.assertEqual(data['result'], expected)

    def test_script_run_async(self):
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        om = self.om
        pkg = 'pkg://{}'.format(pkgpath)
        # put script
        meta = om.scripts.put(pkg, 'helloworld')
        # run the script on the cluster
        resp = self.client.post(self.url('helloworld', action='run', query='text=foo'),
                                headers=self._async_headers)
        resp = self._check_async(resp)
        self.assertHttpOK(resp)
        data = self.deserialize(resp)['response']
        self.assertIn('runtimes', data)
        self.assertIn('result', data)
        expected = list(['hello from helloworld', {'text': 'foo', 'pure_python': False}])
        self.assertEqual(data['result'], expected)



class CombinedOmegaStoreMixin:
    def __init__(self, stores):
        self._stores = stores

    def get(self, name, *args, **kwargs):
        store = self._get_store(name)
        return store.get(name.replace(store.prefix, ''), *args, **kwargs)

    def put(self, obj, name, *args, **kwargs):
        store = self._get_store(name)
        return store.put(obj, name.replace(store.prefix, ''), *args, **kwargs)

    def list(self, pattern=None, regexp=None, **kwargs):
        index = []
        pattern = pattern or ''
        for store in self._stores:
            index.extend(f'{store.prefix}{name}' for name in store.list(pattern.replace(store.prefix, ''),
                                                                        regexp=(regexp.replace(
                                                                            store.prefix) if regexp else regexp),
                                                                        **kwargs))
        return index

    def metadata(self, name):
        store = self._get_store(name)
        return store.metadata(name.replace(store.prefix, ''))

    def help(self, name_or_obj):
        store = self._get_store(name_or_obj)
        return store.help(name_or_obj)

    def _get_store(self, name):
        for store in self._stores:
            if name.startswith(store.prefix):
                return store
        prefixes = [s.prefix for s in self._stores]
        raise ValueError(f'there is no store for {name}, try prefixing with any of {prefixes}')



import tensorflow as tf

from .tfestimatormodel import TFEstimatorModelBackend, TFEstimatorModel
from .tfkeras import TensorflowKerasBackend
from .tfkerassavedmodel import TensorflowKerasSavedModelBackend
from .tfsavedmodel import TensorflowSavedModelPredictor, TensorflowSavedModelBackend, ServingInput

# compatibility with previous tensorflow versions
FN_MAP = {}
if tf.__version__.startswith('1.'):
    FN_MAP['pandas_input_fn'] = getattr(tf.estimator.inputs, 'pandas_input_fn')
    FN_MAP['numpy_input_fn'] = getattr(tf.estimator.inputs, 'numpy_input_fn')
    FN_MAP['convert_to_tensor']  = getattr(tf.compat.v1, 'convert_to_tensor')
elif tf.__version__.startswith('2.'):
    FN_MAP['pandas_input_fn'] = getattr(tf.compat.v1.estimator.inputs, 'pandas_input_fn')
    FN_MAP['numpy_input_fn'] = getattr(tf.compat.v1.estimator.inputs, 'numpy_input_fn')
    FN_MAP['convert_to_tensor'] = getattr(tf, 'convert_to_tensor')


def _tffn(name):
    return FN_MAP[name]



from __future__ import absolute_import

from datetime import timedelta
from unittest import TestCase

import gridfs
from nbformat import v4

from omegaml import Omega
from omegaml.documents import Metadata
from omegaml.notebook.jobs import JobSchedule
from omegaml.util import settings as omegaml_settings, settings


class JobTests(TestCase):

    def setUp(self):
        super().setUp()
        for omx in (self.om, self.om['bucket']):
            for fn in omx.jobs.list():
                omx.jobs.drop(fn)

    @property
    def om(self):
        om = Omega(defaults=settings(reload=True))
        return om

    @property
    def fs(self):
        om = self.om
        defaults = omegaml_settings()
        fs = om.jobs.get_fs(defaults.OMEGA_NOTEBOOK_COLLECTION)
        return fs

    def test_job_put_get(self):
        """
        test job put and get
        """
        om = self.om
        # create a notebook
        cells = []
        code = "print 'hello'"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        # read it back and see what's in it
        notebook2 = om.jobs.get('testjob')
        self.assertDictEqual(notebook2, notebook)

    def test_job_list(self):
        """
        test job listing
        """
        fs = self.fs
        om = self.om
        # create a notebook
        cells = []
        code = "print 'hello'"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        nb = v4.new_notebook(cells=cells)
        job_list = self.om.jobs.list()
        expected = 'testjob.ipynb'
        self.assertIn(expected, job_list)

    def test_job_list_bucket(self):
        """
        test job listing in buckets
        """
        om = self.om
        omb = self.om['bucket']
        # create a notebook
        cells = []
        code = "print 'hello'"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        nb = v4.new_notebook(cells=cells)
        job_list = self.om.jobs.list()
        expected = 'testjob.ipynb'
        # ensure only in default bucket
        self.assertIn(expected, job_list)
        self.assertNotIn(expected, omb.jobs.list())
        # put to new bucket
        omb.jobs.put(notebook, 'testjob')
        self.assertIn(expected, omb.jobs.list())


    def test_run_job_valid(self):
        """
        test running a valid job
        """
        om = self.om
        # create a notebook
        cells = []
        code = "print('hello')"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        meta_job = om.jobs.run('testjob')
        self.assertIsInstance(meta_job, Metadata)
        meta = om.jobs.put(notebook, 'testjob')
        self.assertIn('job_results', meta.attributes)
        self.assertIn('job_runs', meta.attributes)
        runs = meta.attributes['job_runs']
        results = meta.attributes['job_results']
        self.assertEqual(len(runs), 1)
        self.assertEqual(len(results), 1)
        resultnb = results[0]
        self.assertTrue(om.jobs.exists(resultnb))
        self.assertEqual(runs[0]['results'], resultnb)

    def test_run_job_timeout(self):
        """
        test running a job that times out
        """
        om = self.om
        # create a long-running notebook
        cells = []
        code = "import time; time.sleep(15)"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        meta = om.jobs.put(notebook, 'testjob')
        # -- execute with default timeout, expect to succeed
        meta_job = om.jobs.run('testjob')
        self.assertIsInstance(meta_job, Metadata)
        meta = om.jobs.metadata('testjob')
        runs = meta.attributes['job_runs']
        results = meta.attributes['job_results']
        self.assertIn('job_runs', meta.attributes)
        self.assertEqual(len(results), 1)
        resultnb = results[0]
        self.assertTrue(om.jobs.exists(resultnb))
        self.assertEqual(runs[0]['results'], resultnb)
        # -- put the notebook with a timeout less than expected running time
        # -- expect run to fail due to timeout
        om.jobs.drop('testjob', force=True)
        meta = om.jobs.put(notebook, 'testjob')
        meta.kind_meta['ep_kwargs'] = dict(timeout=5)
        meta.save()
        self.assertEqual(meta.name, 'testjob.ipynb')
        meta_job = om.jobs.run('testjob')
        self.assertIsInstance(meta_job, Metadata)
        meta = om.jobs.metadata('testjob')
        self.assertIn('job_runs', meta.attributes)
        runs = meta.attributes['job_runs']
        this_run = runs[0]
        self.assertEqual(this_run['status'], 'ERROR')
        self.assertIn('execution timed out', this_run['message'])
        self.assertEqual(len(runs), 1)
        # -- retry with no timeout
        om.jobs.drop('testjob', force=True)
        meta = om.jobs.put(notebook, 'testjob')
        meta.kind_meta['ep_kwargs'] = dict(timeout=None)
        meta.save()
        meta_job = om.jobs.run('testjob')
        self.assertIsInstance(meta_job, Metadata)
        meta = om.jobs.metadata('testjob')
        runs = meta.attributes['job_runs']
        results = meta.attributes['job_results']
        self.assertIn('job_runs', meta.attributes)
        self.assertEqual(len(results), 1)
        resultnb = results[0]
        self.assertTrue(om.jobs.exists(resultnb))
        self.assertEqual(runs[0]['results'], resultnb)

    def test_run_job_invalid(self):
        """
        test running an invalid job
        """
        fs = self.fs
        om = self.om
        # create a notebook
        cells = []
        code = "INVALID PYTHON CODE"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        nb = v4.new_notebook(cells=cells)
        meta_job = om.jobs.run('testjob')
        self.assertIsInstance(meta_job, Metadata)
        meta = om.jobs.put(notebook, 'testjob')
        runs = meta.attributes['job_runs']
        self.assertEqual(len(runs), 1)
        self.assertEqual('ERROR', runs[0]['status'])

    def test_run_nonexistent_job(self):
        om = self.om
        self.assertRaises(
            gridfs.errors.NoFile, om.jobs.run_notebook, 'dummys.ipynb')

    def test_scheduled_job_with_omegaml_block(self):
        om = self.om
        cells = []
        conf = """
        # omega-ml:
        #   run-at: "*/5 * * * *"
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        meta = om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_job_with_run_at_schedule(self):
        om = self.om
        cells = []
        conf = """
        # run-at: "*/5 * * * *"
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        meta = om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_job_with_cron_schedule(self):
        om = self.om
        cells = []
        conf = """
        # run-at: "*/5 * * * *"
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        meta = om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_job_with_nlp_schedule(self):
        om = self.om
        cells = []
        conf = """
        # schedule: daily, at 06:00
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_not_duplicated(self):
        om = self.om
        cells = []
        conf = """
                # schedule: daily, at 06:00
                """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job(autorun=False, reschedule=False)
        meta = om.jobs.metadata('testjob')
        trigger = meta.attributes['triggers']
        self.assertEqual(len(trigger), 1)

    def test_scheduled_results_not_rescheduled(self):
        om = self.om
        cells = []
        conf = """
                        # schedule: daily, at 06:00
                        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job(autorun=True, reschedule=False)
        meta = om.jobs.metadata('testjob')
        runs = meta.attributes['job_runs']
        result_name = runs[-1]['results']
        meta_result = om.jobs.metadata(result_name)
        self.assertNotIn('triggers', meta_result.attributes)

    def _check_scheduled_job(self, autorun=True, reschedule=False):
        om = self.om
        meta = om.jobs.metadata('testjob')
        config = om.jobs.get_notebook_config('testjob')
        self.assertIn('run-at', config)
        self.assertIn('config', meta.attributes)
        self.assertIn('run-at', meta.attributes['config'])
        # check the job was scheduled
        self.assertIn('triggers', meta.attributes)
        trigger = meta.attributes['triggers'][-1]
        self.assertEqual(trigger['status'], 'PENDING')

        # run it as scheduled, check it is ok
        def get_trigger(event=None):
            # get last trigger or specified by event
            meta = om.jobs.metadata('testjob')
            triggers = meta.attributes['triggers']
            if not event:
                trigger = triggers[-1]
            else:
                trigger = [t for t in triggers if t['event'] == event][0]
            return trigger

        def assert_pending(event=None):
            trigger = get_trigger(event)
            self.assertEqual(trigger['status'], 'PENDING')

        def assert_ok(event=None):
            trigger = get_trigger(event)
            self.assertEqual(trigger['status'], 'OK')

        assert_pending()
        if autorun:
            # -- run by the periodic task. note we pass now= as to simulate a time
            kwargs = dict(now=trigger['run-at'])
            om.runtime.task('omegaml.notebook.tasks.execute_scripts').apply_async(kwargs=kwargs).get()
            assert_ok(event=trigger['event'])
            # -- it should be pending again
            assert_pending()
            # execute the scheduled job by event name
            trigger = get_trigger()
            om.runtime.job('testjob').run(event=trigger['event'])
            # the last run should be ok, and there should not be a new pending event
            # since we did not reschedule
            assert_ok()
            with self.assertRaises(AssertionError):
                assert_pending()
        if reschedule:
            # reschedule explicit and check we have a pending event
            next_time = trigger['run-at'] + timedelta(minutes=2)
            om.runtime.job('testjob').schedule(run_at='daily, at 07:00').get()
            assert_pending()

    def test_schedule_triggers_by_api(self):
        om = self.om
        cells = []
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        om.jobs.put(notebook, 'testjob')
        schedule = om.jobs.Schedule(weekday='mon-fri', at='06:00')
        om.jobs.schedule('testjob', run_at=schedule)

    def test_jobschedule_maker(self):
        # basics
        sched = JobSchedule(minute='*')
        self.assertEqual(sched.text, 'Every minute')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, 'Every minute')
        # day specs
        sched = JobSchedule(weekday='mon-fri', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # day specs 2
        sched = JobSchedule(weekday='Mon-Fri', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # day specs 3
        sched = JobSchedule(weekday='Mon-FRI', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # day specs 4
        sched = JobSchedule(weekday='Mon-FRI', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # multiple times
        sched = JobSchedule(weekday='mon-fri', at='06:05,12:05')
        self.assertEqual(sched.text, 'At 06:05 AM and 12:05 PM, Monday through Friday')
        self.assertEqual(sched.cron, '05 06,12 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # months and days with every
        sched = JobSchedule(month='every 2', at='08:00', weekday='every 3')
        self.assertEqual(sched.text, 'At 08:00 AM, every 3 days of the week, every 2 months')
        self.assertEqual(sched.cron, '00 08 * */2 */3')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # step days
        sched = JobSchedule(weekday='every 2nd', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, every 2 days of the week')
        self.assertEqual(sched.cron, '00 06 * * */2')
        sched = JobSchedule(weekday='every 1st', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Monday')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        sched = JobSchedule(weekday='every 3rd', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, every 3 days of the week')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        sched = JobSchedule(weekday='every 4th', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, every 4 days of the week')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # step hours
        sched = JobSchedule(hour='every 2nd', minute=0, weekday='mon-fri')
        self.assertEqual(sched.text, 'Every 2 hours, Monday through Friday')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # step minutes
        sched = JobSchedule(minute='every 5', weekday='mon-fri')
        self.assertEqual(sched.text, 'Every 5 minutes, Monday through Friday')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # text specs
        sched = JobSchedule('friday, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Friday')
        sched = JobSchedule('fridays, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Friday')
        sched = JobSchedule('Mondays and Fridays, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Monday and Friday')
        sched = JobSchedule('Mondays/Fridays, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Monday and Friday')
        sched = JobSchedule('monday-friday, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('mon-fri, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('mon-fri at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('Mon-Fri, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('Mon-Fri, 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('Mon-Fri 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('at 06:00, Mon-Fri')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('monday-friday, 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 2nd month, monday-friday, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday, every 2 months')
        sched = JobSchedule('every 5 minutes, every day, hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM')
        sched = JobSchedule('every 5 minutes every working day hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes every working day, hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes, every working day, hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes, on workdays, hours 6/7')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM and 07:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes, on workdays, in april')
        self.assertEqual(sched.text, 'Every 5 minutes, Monday through Friday, only in April')
        sched = JobSchedule('every 5 minutes, on weekends, in april')
        self.assertEqual(sched.text, 'Every 5 minutes, Saturday through Sunday, only in April')
        sched = JobSchedule('every 5 minutes, from monday to friday, in april')
        self.assertEqual(sched.text, 'Every 5 minutes, Monday through Friday, only in April')
        sched = JobSchedule('at 5 minutes, every hour, monday to friday, april')
        self.assertEqual(sched.text, 'At 5 minutes past the hour, Monday through Friday, only in April')
        sched = JobSchedule('1st day of month, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, on day 1 of the month')
        sched = JobSchedule('mon-fri, 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 2nd hour, 5 minute, weekdays')
        self.assertEqual(sched.text, 'At 5 minutes past the hour, every 2 hours, Monday through Friday')



import six

from omegaml.backends.basedata import BaseDataBackend
import pandas as pd


class PandasExternalData(BaseDataBackend):

    """
    external data is stored as a uri on some external medium
    """
    SCHEMES = ['http', 'https', 'ftp', 'file', 's3']
    KIND = 'pandas.csv'

    @classmethod
    def supports(cls, obj, name, **kwargs):
        supported_scheme = _is_supported_scheme(obj, **kwargs)
        supported_type = _is_supported_type(obj, **kwargs)
        return supported_scheme | supported_type

    def put(self, obj, name, attributes=None, **kwargs):
        """
        Store external data (currently only csv)

        This supports various pandas file sources as:

        * storing the URI to an external csv file
        * storing a pd.DataFrame, pd.Series, as a csv, at an external URI
        * copying an external CSV from one URI to another URI

        Only URIs implemented by pandas.read_csv() are supported to store
        the URI (http, ftp, file, s3). Only s3 and local files are supported
        to copy a csv file from either a supported URI or from an object
        into a csv on s3 or on the local file system.

        Examples:

        # store a reference to S3
        om.datasets.put('s3://bucket/test/google.csv', 'google-s3')
        # from dataframe direcly to s3
        om.datasets.put(df, 'google-s3', uri='s3://bucket/test/google.csv')
        # from s3 into mongodb
        om.datasets.put(None, 'google-mongo', uri='s3://bucket/test/google.csv')
        # from dataframe to local
        om.datasets.put(df, 'google-local', uri='file:///tmp/google.csv')
        # from local directly to s3
        om.datasets.put('s3://bucket/test/google2.csv',
        'google-s3-direct', uri='file:///tmp/google.csv')

        :param obj: (str|pd.Series|pd.DataFrame) the object to store. If str
          must be URI of the form scheme://location/name.csv where scheme
          is one of the upported schemes by pandas.read_csv
        :param name: (str) the name of the object in the store
        :param attributes: custom attributes to store with the object
        """
        uri = kwargs.get('uri')
        # case object is a supported uri, e.g. s3://bucket/name.ext
        if _is_supported_scheme(obj, **kwargs):
            if uri is None:
                # store object as a location pointer with given name
                metadata = self.data_store.make_metadata(name, self.KIND,
                                                         attributes=attributes)
                metadata.uri = obj
            else:
                # read object, store in given location
                # then store as a location pointer with given name
                df = pd.read_csv(uri)
                metadata = self.data_store.put(df, name, uri=obj,
                                               attributes=attributes)
        # case object is a supported object type
        elif _is_supported_type(obj, **kwargs) and uri:
            # if object is None we read the file from the URI first
            if obj is None:
                obj = pd.read_csv(uri)
            # convert to CSV data
            data = obj.to_csv()
            # store in supported location
            if uri.startswith('s3://'):
                import s3fs
                data = obj.to_csv()
                fs = s3fs.S3FileSystem()
                with fs.open(uri, 'wb') as fout:
                    fout.write(data.encode(encoding='utf-8'))
                metadata = self.data_store.make_metadata(
                    name, self.KIND, attributes=attributes)
                metadata.uri = uri
            elif uri.startswith('file'):
                obj.to_csv(uri.replace('file://', ''))
                metadata = self.data_store.make_metadata(
                    name, self.KIND, attributes=attributes)
                metadata.uri = uri
            # case erroneous
            else:
                raise ValueError(
                    ('Not supported type obj={} with '
                     'name={} and uri={}').format(type(obj), name, uri))
        # case erroneous parameters
        else:
            raise ValueError(('Not supported type obj={} '
                              'with name={}').format(type(obj), name))
        return metadata.save()

    def get(self, name, version=-1, force_python=False, lazy=False, **kwargs):
        """
        Get a csv stored at an external location
        """
        meta = self.data_store.metadata(name)
        read_kwargs = meta.kind_meta.get('pandas_kwargs', {})
        read_kwargs.update(dict(usecols=kwargs.get('columns')))
        read_kwargs.update(kwargs)
        read_kwargs.update(kwargs.get('pandas_kwargs', {}))
        df = pd.read_csv(meta.uri, **read_kwargs)
        return df


def _is_supported_scheme(obj, **kwargs):
    if isinstance(obj, six.string_types):
        parts = obj.split('://')
        if len(parts) > 1:
            scheme, loc = parts
            return scheme in PandasExternalData.SCHEMES
    elif obj is None and kwargs.get('uri') is not None:
        return True
    return False


def _is_supported_type(obj, **kwargs):
    uri = kwargs.get('uri')
    supported_type = isinstance(obj, (pd.Series, pd.DataFrame))
    supported_scheme = _is_supported_scheme(uri)
    return supported_type & supported_scheme




class GenericScriptResource(object):
    def __init__(self, om, is_async=False):
        self.om = om
        self.is_async = is_async

    def is_eager(self):
        return getattr(self.om.runtime.celeryapp.conf, 'CELERY_ALWAYS_EAGER', False)

    def run(self, script_id, query, payload):
        """
        Args:
            script_id (str): the name of the model
            query (dict): the query parameters
            payload (dict): the json body

        Returns:
            result
        """
        om = self.om
        payload = {} if payload is None else payload
        promise = om.runtime.script(script_id).run(__format='python', **query, **payload)
        result = self.prepare_result_from_run(promise.get(), script_id=script_id) if not self.is_async else promise
        return result

    def prepare_result_from_run(self, result, script_id=None, **kwargs):
        return result



class GenericJobResource(object):
    def __init__(self, om, is_async=False):
        self.om = om
        self.is_async = is_async

    def is_eager(self):
        return getattr(self.om.runtime.celeryapp.conf, 'CELERY_ALWAYS_EAGER', False)

    def run(self, job_id, query, payload):
        """
        Args:
            job_id (str): the name of the model
            query (dict): the query parameters
            payload (dict): the json body

        Returns:
            dict(job: id, result: dict)
        """
        om = self.om
        promise = om.runtime.job(job_id).run()
        result = self.prepare_result_from_run(promise.get(), job_id=job_id) if not self.is_async else promise
        return result

    def metadata(self, job_id, query, payload):
        return self._get_job_detail(job_id)

    def prepare_result_from_run(self, result, job_id=None, **kwargs):
        return self._get_job_detail(job_id)

    def _get_job_detail(self, job_id):
        meta = self.om.jobs.metadata(job_id)
        data = {
            'job': meta.name,
            'job_results': meta.attributes.get('job_results', {}),
            'job_runs': meta.attributes.get('job_runs', []),
            'created': meta.created,
        }
        if 'source_job' in meta.attributes:
            data['source_job'] = meta.attributes['source_job']
        return data



from joblib import Parallel, delayed

from omegaml.util import PickableCollection


class ParallelApplyMixin:
    """
    Enables parallel apply to MDataFrame

    This enables Python-native Pandas processing to be applied to MDataFrames
    of any size

    Usage:
        def myfunc(df):
            # this function is executed for each chunk in parallel
            # use any Pandas DataFrame native functions
            df[...] = df.apply(...)

        om.datasets.getl('verylarge').transform(myfunc).persist('transformed')
    """

    def transform(self, fn=None, n_jobs=-2, maxobs=None,
                  chunksize=50000, chunkfn=None, outname=None,
                  resolve='worker', backend='omegaml'):
        """

        Args:
            fn (func): the function to apply to each chunk, receives the
               chunk DataFrame (resolve==worker) or MDataFrame (resolve==function)
            n_jobs (int): number of jobs, defaults to CPU count - 1
            maxobs (int): number of max observations to process, defaults to
               len of mdf
            chunksize (int): max size of each chunk, defaults to 50000
            chunkfn (func): the function to chunk by
            outname (name): output collection name, defaults to _tmp_ prefix of
              input name
            resolve (string): worker, function. If worker is specified chunkfn
                receives the resolved DataFrame, otherwise receives the MDataFrame.
                Specify function to apply a custom resolving strategy (e.g. processing
                records one by one). Defaults to worker which uses .value on each
                chunk to resolve each chunk to a DataFrame before sending

        See Also:
            https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html

        Returns:

        """
        mdf = self.__class__(self.collection, **self._getcopy_kwargs())
        options = mdf._transform_options()
        options.update({
            'maxobs': maxobs or len(mdf),
            'n_jobs': n_jobs,
            'chunksize': chunksize,
            'applyfn': fn or pyappply_nop_transform,
            'chunkfn': chunkfn,
            'mdf': mdf,
            'append': False,
            'outname': outname or '_tmp{}_'.format(mdf.collection.name),
            'resolve': resolve,  # worker or function
            'backend': backend,
        })
        return mdf

    def _chunker(self, mdf, chunksize, maxobs):
        if getattr(mdf.collection, 'query', None):
            for i in range(0, maxobs, chunksize):
                yield mdf.skip(i).head(i + chunksize)
        else:
            for i in range(0, maxobs, chunksize):
                yield mdf.iloc[i:i + chunksize]

    def _do_transform(self, verbose=0):
        # setup mdf and parameters
        opts = self._transform_options()
        n_jobs = opts['n_jobs']
        chunksize = opts['chunksize']
        applyfn = opts['applyfn']
        chunkfn = opts['chunkfn'] or self._chunker
        maxobs = opts['maxobs']
        mdf = opts['mdf']
        outname = opts['outname']
        append = opts['append']
        resolve = opts['resolve']
        backend = opts['backend']
        outcoll = PickableCollection(mdf.collection.database[outname])
        if not append:
            outcoll.drop()
        non_transforming = lambda mdf: mdf._clone()
        with Parallel(n_jobs=n_jobs, backend=backend,
                      verbose=verbose) as p:
            # prepare for serialization to remote worker
            chunks = chunkfn(non_transforming(mdf), chunksize, maxobs)
            runner = delayed(pyapply_process_chunk)
            worker_resolves_mdf = resolve in ('worker', 'w')
            # run in parallel
            jobs = [runner(mdf, i, chunksize, applyfn, outcoll, worker_resolves_mdf)
                    for i, mdf in enumerate(chunks)]
            p._backend._job_count = len(jobs)
            if verbose:
                print("Submitting {} tasks".format(len(jobs)))
            p(jobs)
        return outcoll

    def _get_cursor(self, pipeline=None, use_cache=True):
        # called by .value
        if self._transform_options():
            result = self._do_transform().find()
        else:
            result = super()._get_cursor(pipeline=pipeline, use_cache=use_cache)
        return result

    @property
    def is_transforming(self):
        return bool(self._transform_options())

    def _transform_options(self):
        self._pyapply_opts = getattr(self, '_pyapply_opts', {})
        return self._pyapply_opts

    def persist(self, name=None, store=None, append=False, local=False):
        """
        Evaluate and persist the result of a .transform() in chunks

        Args:
            name (str): the name of the target dataset
            store (OmegaStore): the target store, defaults to om.datasets
            append (bool): if True will append the data, otherwise replace. Defaults
               to False
            local (bool): if True resolves a pending aggregation result into memory first, then persists
               result. Defaults to False, effectively persisting the result by the database without
               returning to the local process

        Returns:
            Metadata of persisted dataset
        """
        # -- .transform() active
        options = self._transform_options()
        if options:
            meta = None
            if name and store:
                coll = store.collection(name)
                if coll.name == self.collection.name:
                    raise ValueError('persist() must be to a different collection than already existing')
                try:
                    if not append:
                        store.drop(name, force=True)
                    meta = store.put(coll, name)
                except:
                    print("WARNING please upgrade omegaml to support accessing collections")
                else:
                    # _do_transform expects the collection name, not the store's name
                    name = coll.name
            options.update(dict(outname=name, append=append))
            coll = self._do_transform()
            result = meta or self.__class__(coll, **self._getcopy_kwargs())
        # -- run with noop in parallel
        elif not local and getattr(self, 'apply_fn', None) is None and name:
            # convenience, instead of .value call mdf.persist('name', store=om.datasets)
            result = self.transform().persist(name=name, store=store, append=append)
        # -- some other action is active, e.g. groupby, apply
        elif local or (name and store):
            print("warning: resolving the result of aggregation locally before storing")
            value = self.value
            result = store.put(value, name, append=append)
        else:
            result = super().persist()
        return result


def pyappply_nop_transform(ldf):
    # default transform that does no transformations
    pass


def pyapply_process_chunk(mdf, i, chunksize, applyfn, outcoll, worker_resolves):
    # chunk processor
    import pandas as pd
    from inspect import signature
    # fix pickling issues
    mdf._parser = getattr(mdf, '_parser', None)
    mdf._raw = getattr(mdf, '_raw', None)
    # check apply fn so we can pass the right number of args
    sig = signature(applyfn)
    params = sig.parameters
    try:
        if worker_resolves:
            # requested to resolve value before passing on
            chunkdf = mdf.value
        else:
            # requested to pass on mdf itself
            chunkdf = mdf
    except Exception as e:
        raise e
        raise RuntimeError(f".value on {mdf} cause exception {e})")
    else:
        applyfn_args = [chunkdf, i][0:len(params)]
    # call applyfn
    if len(chunkdf):
        try:
            result = applyfn(*applyfn_args)
        except Exception as e:
            raise RuntimeError(e)
        else:
            chunkdf = result if result is not None else chunkdf
            if isinstance(chunkdf, dict):
                chunkdf = pd.DataFrame(chunkdf)
            if isinstance(chunkdf, pd.Series):
                chunkdf = pd.DataFrame(chunkdf,
                                       index=chunkdf.index,
                                       columns=[str(chunkdf.name)])
        start = i * chunksize
        if chunkdf is not None and len(chunkdf):
            end = start + len(chunkdf)
            chunkdf['_om#rowid'] = pd.RangeIndex(start, end)
            outcoll.insert_many(chunkdf.to_dict(orient='records'))



from warnings import warn

import os
import six
from mongoengine import GridFSProxy


class BackendBaseCommon:
    """
    common base for storage backends
    """
    def _tmp_packagefn(self, store, name):
        """
        use this to to get a temporary local filename for serialization/deserialization

        This uses the store's .tmppath to generate a valid fully qualified path, ensuring
        the directory exists. Note that you must clean up files yourself after use, i.e.
        call os.remove()

        Args:
            store (OmegaStore): the store to get the temporary path
            name (str): the name of the file

        Returns:
            filename (str)

        TODO: register all files to remove temp files on process exit
        """
        filename = os.path.join(store.tmppath, name)
        dirname = os.path.dirname(filename)
        os.makedirs(dirname, exist_ok=True)
        return filename

    def _is_path(self, obj):
        return isinstance(obj, six.string_types) and os.path.exists(obj)

    def _store_to_file(self, store, obj, filename, encoding=None, replace=False):
        """
        Use this method to store file-like objects to the store's gridfs

        Args:
            store (OmegaStore): the store whose .fs filesystem access will be used
            obj (file-like): a file-like object or path, if path will be opened with mode=rb,
               otherwise its obj.read() method is called to get the data as a byte stream
            filename (path): the path in the store (key)
            encoding (str): a valid encoding such as utf8, optional
            replace (bool): if True the existing file(s) of the same name are deleted to avoid automated versioning
               by gridfs. defaults to False

        Returns:
            gridfile (GridFSProxy), assignable to Metadata.gridfile
        """
        if replace:
            for fileobj in store.fs.find({'filename': filename}):
                try:
                    store.fs.delete(fileobj._id)
                except Exception as e:
                    warn('deleting {filename} resulted in {e}'.format(**locals()))
                    pass
        if self._is_path(obj):
            with open(obj, 'rb') as fin:
                fileid = store.fs.put(fin, filename=filename, encoding=encoding)
        else:
            fileid = store.fs.put(obj, filename=filename, encoding=encoding)
        gridfile = GridFSProxy(grid_id=fileid,
                               db_alias=store._dbalias,
                               key=filename,
                               collection_name=store._fs_collection)
        return gridfile




from multiprocessing import Process, Queue
from unittest import TestCase

from minibatch import Buffer, connectdb, stream
from minibatch.contrib.omegaml import DatasetSink
from time import sleep

from omegaml import Omega
from omegaml.util import delete_database


class MiniBatchTests(TestCase):
    def setUp(self):
        delete_database()
        self.om = Omega()
        db = self.om.datasets.mongodb
        self.url = self.om.mongo_url + '?authSource=admin'
        connectdb(url=self.url)

    def test_stream(self):
        """
        Test a stream writes to a buffer
        """
        from minibatch import stream
        om = self.om
        om.datasets.mongodb
        s = stream('test', url=self.url)
        s.append({'foo': 'bar1'})
        s.append({'foo': 'bar2'})
        count = len(list(doc for doc in Buffer.objects.all()))
        self.assertEqual(count, 2)

    def test_fixed_size(self):
        """
        Test batch windows of fixed sizes work ok
        """
        from minibatch import streaming, stream

        def consumer(q, url):
            # note the stream decorator blocks the consumer and runs the decorated
            # function asynchronously upon the window criteria is satisfied
            om = Omega(mongo_url=url)

            @streaming('test', size=2, url=url, keep=True, queue=q,
                       sink=DatasetSink(om, 'consumer'))
            def myprocess(window):
                return {'myprocess': True, 'data': window.data}

        # start stream and consumer
        s = stream('test', url=self.url)
        q = Queue()
        proc = Process(target=consumer, args=(q, self.url))
        proc.start()
        # fill stream
        for i in range(10):
            s.append({'index': i})
        # give it some time to process
        sleep(5)
        q.put(True)
        proc.join()
        # expect 5 entries, each of length 2
        windows = list(doc for doc in self.om.datasets.collection('consumer').find())
        self.assertEqual(len(windows), 5)
        self.assertTrue(all(len(w['data']) == 2 for w in windows))

    def test_timed_window(self):
        """
        Test timed windows work ok
        """
        from minibatch import streaming

        def consumer(q, url):
            # note the stream decorator blocks the consumer and runs the decorated
            # function asynchronously upon the window criteria is satisfied
            om = Omega(mongo_url=url)

            @streaming('test', interval=1, keep=True, url=url, queue=q,
                       sink=DatasetSink(om, 'consumer'))
            def myprocess(window):
                return {'myprocess': True, 'data': window.data}

        # start stream and consumer
        q = Queue()
        s = stream('test', url=self.url)
        proc = Process(target=consumer, args=(q, self.url,))
        proc.start()
        # fill stream
        for i in range(10):
            s.append({'index': i})
            sleep(.5)
        # give it some time to process
        sleep(5)
        q.put(True)
        proc.join()
        # expect at least 5 entries (10 x .5 = 5 seconds), each of length 1-2
        windows = list(doc for doc in self.om.datasets.collection('consumer').find())
        self.assertGreater(len(windows), 5)
        print(windows)
        self.assertTrue(all(len(w['data']) >= 1 for w in windows))

    def test_timed_window_relaxed(self):
        """
        Test relaxed time windows work ok
        """
        from minibatch import streaming

        def consumer(q, url):
            # note the stream decorator blocks the consumer and runs the decorated
            # function asynchronously upon the window criteria is satisfied
            om = Omega(mongo_url=url)

            @streaming('test', interval=1, relaxed=True,
                       keep=True, url=url, queue=q,
                       sink=DatasetSink(om, 'consumer'))
            def myprocess(window):
                return {'myprocess': True, 'data': window.data}

        # start stream and consumer
        s = stream('test', url=self.url)
        q = Queue()
        proc = Process(target=consumer, args=(q, self.url,))
        proc.start()
        # fill stream
        for i in range(10):
            s.append({'index': i})
            sleep(.5)
        # give it some time to process
        sleep(5)
        q.put(True)
        proc.join()
        # expect at least 5 entries (10 x .5 = 5 seconds), each of length 1-2
        windows = list(doc for doc in self.om.datasets.collection('consumer').find())
        self.assertGreater(len(windows), 5)
        self.assertTrue(all(len(w['data']) >= 1 for w in windows))
